/*
 * Aspera, Inc.
 *
 * Internal class.
 *
 * Used in the TCP out-of-band / tunneling feature. This class's run
 * method is executed by a background thread.
 *
 * Tunnels communication to and from remote faspstream through
 * local socket connected to faspstream process.
 */
#include <Poco/Runnable.h>
#include <Poco/Net/StreamSocket.h>
#include <Poco/Mutex.h>
#include "FaspConnectionChannel.h"

class ChannelHandler : public Poco::Runnable
{
public:
    /**
     * Constructor.
     *
     * @param socket a TCP socket connected to the local faspstream process
     * @param channel the abstract interface to share TCP connection
     */
    ChannelHandler(Poco::Net::StreamSocket& socket,
        FaspConnectionChannel& channel) :
        _socket(socket),
        _channel(&channel),
        _pBuffer(new char[BUFFER_SIZE]),
        _shouldStop(0)
    {
    }
    ~ChannelHandler() {
        delete [] this->_pBuffer;
        _socket.close();
        _channel->close();
    }

    /**
     * Invoked by Poco::Thread when it is spawned. This runs the
     * main loop that tunnels data between the local and remote
     * faspstreams. Part of the Poco::Runnable interface.
     */
    virtual void run();
    /**
     * Call this when the thread should stop. This will close
     * the TCP connection to the local faspstream process and
     * invoke the channel's close method.
     */
    virtual void stop();

private:
    /**
     * Size of buffer used for internal buffering of data
     * tunneled between local and remote faspstreams.
     */
    enum {
        BUFFER_SIZE = 1024
    };

    Poco::Net::StreamSocket  _socket;
    FaspConnectionChannel* _channel;
    Poco::Mutex _mutex;
    char* _pBuffer;
    int _shouldStop;

    /**
     * Private helpers (so we can use instance variables).
     *
     * TODO: Might be better to use static functions and pass
     * ivars as arguments so that we hide these details.
     */
    int try_send_remote(Poco::Net::StreamSocket& ss);
    int try_read_remote(Poco::Net::StreamSocket& ss);
    bool should_stop();

};

