/*
 * Aspera, Inc.
 *
 * C++ only.
 *
 * FaspInputStream class.
 */
#pragma once

#include <string>
#include <vector>

#include <Poco/Process.h>
#include <Poco/SharedPtr.h>
#include <Poco/PipeStream.h>

#include "Constants.h"
#include "FaspStreamLogger.h"

#ifdef WIN32

#ifdef FASPSTREAM_CPP_EXPORTS
#define FASPSTREAM_CPP_API __declspec(dllexport)
#else
#define FASPSTREAM_CPP_API __declspec(dllimport)
#endif

#else // NOT WIN32
#define FASPSTREAM_CPP_API
#endif

/**
 * Use the FaspInputStream class to read faspstream data.  Note this class does
 * not inherit from any of the C++ std::ios stream classes, and
 * requires the user to implement a read() loop to receive the data.
 */
class FaspInputStream
{

public:
    /**
     * Initializes fasp input stream to use outPipe connected
     * to faspstream process. The faspstream process should write
     * to outPipe, and FaspInputStream will read from it.
     */
    FASPSTREAM_CPP_API FaspInputStream(const Poco::Pipe &outPipe,
                                       const Poco::SharedPtr<Poco::ProcessHandle> ph,
                                       const FaspStreamLogger& logger);
    /**
     * Reads available data from faspstream.
     *
     * Returns the number of bytes read, zero for end of file, negative
     * for error.
     *
     * This is a blocking function.
     *
     * Throws a FaspStreamException if there is an error reading from the
     * faspstream.
     *
     * NOTE: You may also encounter a SIGPIPE if the faspstream process
     *       exited prematurely (perhaps due to an error).
     *
     * @param buffer pointer to buffer containing bytes read (out parameter)
     * @param offset number of bytes into the buffer to begin writing data
     * @param len number of bytes read into buf
     * @return number of bytes read, zero for end of file, negative for error
     */
    FASPSTREAM_CPP_API int read(char* buffer, int offset, int length);
    /**
     * Returns true if end of file (stream) was reached while reading.
     *
     * NOTE: read returns 0 to indicate end of file. This method may be
     *       deprecated/removed in the future.
     */
    FASPSTREAM_CPP_API bool eof();
    /**
     * Closes the pipe to the faspstream process.
     */
    FASPSTREAM_CPP_API void close();
    FASPSTREAM_CPP_API void abort();
    FASPSTREAM_CPP_API void cleanup();

private:
    Poco::Pipe _outPipe;
    Poco::SharedPtr<Poco::ProcessHandle> _ph;
    bool _eof;
    FaspStreamLogger _logger;
};
