/*
 * Aspera, Inc.
 *
 * Example client that uses the C wrapper for the FaspStream
 * SDK.
 *
 * Usage:
 *
 *   ./client < outfile
 *
 * Uses mostly default settings. Uses aspera feed.
 *
 * Linux only.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fasp_stream_client.h>
#include <fasp_stream_constants.h>
#include <fasp_stream_channel.h>
#include "../poco_stream_socket/poco_stream_socket.h"
#ifdef WIN32
        #include <io.h>
        #include <fcntl.h>
#endif

/*
 * Configure these for the remote server.
 */
#define REMOTE_HOST "localhost"
#define REMOTE_PORT 4444

static
void
err(char *msg)
{
    fprintf(stderr, "error: client: %s\n", msg);
}

static
void
log_msg(const char *msg)
{
    fprintf(stderr, "%s\n", msg);
}

static
int
poco_read(fasp_stream_socket_wrapper_t *w, char *buff, int len, int *timedout)
{
    int ret;
    int sock_ret;
    poco_stream_socket_t *s;
    /*
     * Pull out poco socket from wrapper
     */
    s = (poco_stream_socket_t *)w->priv;
    /*
     * Do read on poco socket
     */
    ret = poco_stream_socket_receive_bytes(s, buff, len, 0,
        &sock_ret, timedout);
    if (ret < 0)
        return ret;

    return sock_ret;
}


static
int
poco_write(fasp_stream_socket_wrapper_t *w, const char *buff, int len)
{
    int ret;
    int sock_ret;
    poco_stream_socket_t *s;
    /*
     * Pull out poco socket from wrapper
     */
    s = (poco_stream_socket_t *)w->priv;
    /*
     * Do write on socket
     */
    ret = poco_stream_socket_send_bytes(s, buff, len, 0, &sock_ret);
    if (ret < 0)
        return ret;

    return sock_ret;
}

static
void
poco_close(fasp_stream_socket_wrapper_t *w)
{
    /*
     * Don't do anything (we'll close it in main).
     */
}

static
int
setup_wrapper(fasp_stream_socket_wrapper_t *sock_wrapper)
{
    int ret;
    poco_stream_socket_t *sock;
    /*
     * Connect to server
     */
    ret = poco_stream_socket_client(REMOTE_HOST, REMOTE_PORT, &sock);
    if (ret) {
        err("error connecting to server");
        goto fail1;
    }
    /*
     * Set timeout on Poco socket
     */
    poco_stream_socket_set_receive_timeout(sock, FASPSTREAM_CHANNEL_TIMEOUT);
    /*
     * Set up channel
     */
    sock_wrapper->priv = sock;
    sock_wrapper->read = poco_read;
    sock_wrapper->write = poco_write;
    sock_wrapper->close = poco_close;

    return 0;

fail1:
    return ret;
}

static
int
setup_fasp_stream(fasp_stream_socket_wrapper_t *wrapper,
    fasp_stream_client_t **out)
{

    fasp_stream_client_t *fsc;
    int ret;
    /*
     * Create the client
     */
    ret = fasp_stream_channel_client_new("client-id", &fsc, wrapper);
    if (ret) {
        err("creating client");
        goto fail;
    }
    /*
     * Configure client
     * ----------------
     *
     * NOTE: TCP port is not needed in configuration since we are
     * using out-of-band / tunneling.
     *
     * Path to faspstream binary (not the SDK library).
     */
    fasp_stream_client_set_faspstream_path(fsc, FASPSTREAM_BINARY);
    /*
     * Set transfer mode to SEND (client will send data to server)
     */
    fasp_stream_client_set_transfer_mode(fsc,
        FASPSTREAM_TRANSFER_MODE_TYPE_DIRECT_SEND);
    /*
     * Rate target, in Kbps (100 Mbps)
     */
    fasp_stream_client_set_rate_target(fsc, 100000);
    /*
     * Flush internal buffer after 1 second.
     */
    fasp_stream_client_set_force_send_milliseconds(fsc, 1000);
    /*
     * Set UDP port for the fasp transfer
     */
    fasp_stream_client_set_udp_port(fsc, 33001);

    /*
     * Register logging function callback
     */
    fasp_stream_client_register_log_callback(fsc, &log_msg);

    *out = fsc;

    return 0;

fail:
    return ret;
}

static
int
read_bytes(char *buff, int len)
{
    int ret;
    /*
     * Read bytes from stdin
     */
    ret = fread(buff, 1, len, stdin);
    if (ret < 0) {
        err("trouble reading from stdin?");
        return -1;
    }
    if (ret == 0 && feof(stdin)) {
        return 0;
    }

    return ret;
}

int
main() {
    fasp_stream_client_t *fsc;
    fasp_stream_socket_wrapper_t wrapper;
    int ret;
    int count;
    int total_bytes_sent;
    char buff[512];
#ifdef WIN32
    /*
     * Required on Windows for binary files.
     */
    _setmode(_fileno(stdin), O_BINARY);
#endif
    /*
     * Connect to server, and set up channel for tunneling.
     */
    fprintf(stderr, "Connecting to server...\n");
    ret = setup_wrapper(&wrapper);
    if (ret)
        goto fail0;
    /*
     * Create and set up faspstream client
     */
    fprintf(stderr, "Connected to server, setting up tunneling...\n");
    ret = setup_fasp_stream(&wrapper, &fsc);
    if (ret)
        goto fail1;
    /*
     * Fire up faspstream and tunneling. Host should be the remote
     * peer where the remote faspstream is running. This is most likely
     * the same host/ip that you used to create the regular TCP socket.
     */
    ret = fasp_stream_client_connect(fsc, REMOTE_HOST);
    if (ret) {
        err("error connecting to server");
        goto fail2;
    }
    fprintf(stderr, "Tunneling set up, sending data...\n");
    /*
     * Send data
     */
    total_bytes_sent = 0;
    for (;;) {
        /*
         * Read bytes from stdin
         */
        count = read_bytes(buff, sizeof(buff));
        if (count <= 0) {
            /*
             * EOF or error
             */
            break;
        }
        /*
         * Write to client's output stream. All bytes are written unless
         * there's an error.
         */
        ret = fasp_stream_client_write(fsc, buff, count);
        if (ret) {
            err("error writing to client's output stream");
            goto fail2;
        }
        /*
         * Update progress
         */
        total_bytes_sent += count;
        fprintf(stderr, "\rBytes Sent: %d", total_bytes_sent);
        fflush(stderr);
    }

    ret = count;

    /*
     * Clean up
     */
    ret = fasp_stream_client_cleanup(fsc);
    if (ret) {
        err("error cleaning up client");
        goto fail2;
    }

    fprintf(stderr, "\n");

    goto out;

out:
fail2:
    /*
     * Destroy client
     */
    if (fasp_stream_client_destroy(fsc)) {
        err("error destroying client");
        abort();
    }
fail1:
    /*
     * Destroy poco socket
     */
    poco_stream_socket_destroy((poco_stream_socket_t *)wrapper.priv);
fail0:
    return ret;
}
