/*
 * Aspera, Inc.
 *
 * Server.cpp
 *
 * Shows how to listen for an incoming client connection and receive
 * a stream of bytes from it and write them to stdout.
 *
 * Usage:
 *
 *    ./Server > out.file
 */
#include <iostream>
#include <fstream>
#include <Constants.h>
#include <FaspInputStream.h>
#include <FaspStreamServer.h>
#include <FaspStreamException.h>
#ifdef WIN32
        #include <io.h>
        #include <fcntl.h>
#endif

static
void
setupStreams() {
#ifdef WIN32
    /*
     * Required on Windows for binary files
     */
    _setmode(fileno(stdin), O_BINARY);
    _setmode(fileno(stdout), O_BINARY);
#endif
    std::cin.exceptions(std::ifstream::badbit);
    std::cout.exceptions(std::ifstream::badbit);
}

static
void
log_msg(const char *msg)
{
    fprintf(stderr, "%s\n", msg);
}


int main()
{
    char data[512];
    setupStreams();
    try {
        /*
         * Initialize faspstream server
         */
        FaspStreamServer fss("Unique-ID-For-a-Client");
        /*
         * Configure server
         * ----------------
         *
         * Set target rate to 100 Mbps
         */
        fss.setRateTarget(100000);
        /*
         * Set the port to connect to
         */
        fss.setPort(4444);
        /*
         * Set the UDP Port for the fasp transfer
         */
        fss.setUdpPort(33001);

        /*
        * Flush internal buffer after 1 second.
        */
        fss.setForceSendMilliseconds(1000);

        fss.setTransferMode(FASPSTREAM_TRANSFER_MODE_TYPE_DIRECT_RECEIVE);

        /*
         * Register logging function callback
         */
        fss.registerLogCallback(log_msg);
        /*
         * Start listening for client
         * --------------------------
         */
        std::cerr << "Waiting for client to connect..." << std::endl;
        fss.startServer(true);
        /*
         * Receive data
         */

        long totalCount = 0;
        int count = 0;
        do {
            /*
             * Read from fasp stream
             */
            count = fss.getFaspInputStream().read(data, 0, sizeof(data));
            totalCount += count;
            /*
             * Write to stdout
             */
            std::cout.write(data, count);

            std::cerr << "Bytes Received: " << totalCount << "\r";
            std::cerr.flush();
        } while (count > 0);

        std::cout.flush();
        /*
         * Clean up server
         */
        fss.cleanup();
        std::cerr << std::endl << "Transfer Complete!" << std::endl;
        std::cerr << "Bytes Received: " << totalCount << std::endl;

    }
    catch (std::ofstream::failure &e) {
        std::cerr << "Error writing to stdout: " << std::endl;
        std::cerr << e.what();
    }
    catch (std::exception &e) {
        std::cerr << "A standard exception occurred:" << std::endl;
        std::cerr << e.what();
    }
    catch (...) {
        std::cerr << "An unknown exception ocurred! ";
        std::cerr << "(No further information is available)";
    }

    return 0;
}

