/*
 * Aspera, Inc.
 *
 * Client.cpp
 *
 * Shows how to connect to a faspstream server and *receive* a stream
 * of bytes and write them to stdout.
 *
 * Usage:
 *
 *   ./Client > out.file
 */
#include <iostream>
#include <fstream>
#include <Constants.h>
#include <FaspInputStream.h>
#include <FaspStreamClient.h>
#include <FaspStreamException.h>
#ifdef WIN32
        #include <io.h>
        #include <fcntl.h>
#endif

static
void
setupStreams() {
#ifdef WIN32
    /*
     * Required on Windows for binary files
     */
    _setmode(fileno(stdin), O_BINARY);
    _setmode(fileno(stdout), O_BINARY);
#endif
    std::cin.exceptions(std::ifstream::badbit);
    std::cout.exceptions(std::ifstream::badbit);
}

int main()
{
    char data[512];
    setupStreams();
    try {
        /*
         * Initialize faspstream client
         */
        FaspStreamClient fsc("Unique-ID-For-a-Client");
        /*
         * Configure client
         * ----------------
         *
         * Set target rate to 100 Mbps
         */
        fsc.setRateTarget(100000);
        /*
         * Set the port to connect to
         */
        fsc.setPort(4444);
        /*
         * Set the UDP Port for the fasp transfer
         */
        fsc.setUdpPort(33001);
        /*
         * Set transfer mode to *receive*
         */
        fsc.setTransferMode(FASPSTREAM_TRANSFER_MODE_TYPE_DIRECT_RECEIVE);
        /*
         * Connect to server
         * -----------------
         */
        fsc.connect("localhost");
        std::cerr << "Connected to server, sending data..."
                  << std::endl;
        /*
         * Receive data
         */
        long totalCount = 0;
        int count = 0;
        do {
            /*
             * Read from fasp stream
             */
            count = fsc.getFaspInputStream().read(data, 0, sizeof(data));
            totalCount += count;
            /*
             * Write to stdout
             */
            std::cout.write(data, count);

            std::cerr << "Bytes Received: " << totalCount << "\r";
            std::cerr.flush();
        } while (count > 0);
        
        std::cout.flush();
        /*
         * Done; clean up client.
         */
        fsc.cleanup();

        std::cerr << std::endl 
                  << "Transfer Complete!" 
                  << std::endl;
        std::cerr << "Bytes Received: " 
                  << totalCount
                  << std::endl;

    }
    catch (std::ifstream::failure &e) {
        std::cerr << "Error writing to stdout: "
                  << std::endl
                  << e.what();
    } 
    catch (std::exception &e) {
        std::cerr << "A standard exception occurred:" << std::endl;
        std::cerr << e.what();
    } 
    catch (...)  {
        std::cerr << "An unknown exception ocurred! ";
        std::cerr << "(No further information is available)";
    }

    return 0;
}
