/*
 * Aspera, Inc.
 *
 * Server.cpp
 *
 * Shows how to use a single TCP connection for application-specific
 * data and fasp control data. Receives data from client and writes
 * to stdout.
 *
 * Usage:
 *
 *    ./Server > out.file
 */
#include <iostream>
#include <fstream>

#include <Poco/Net/SocketAddress.h>
#include <Poco/Net/StreamSocket.h>
#include <Poco/Net/NetException.h>
#include <Poco/Net/ServerSocket.h>

#include <Constants.h>
#include <FaspInputStream.h>
#include <FaspStreamServer.h>
#include <FaspStreamException.h>
#ifdef WIN32
        #include <io.h>
        #include <fcntl.h>
#endif

/*
 * To share a TCP connection with faspstream, you must implement
 * the FaspConnectionChannel interface, in FaspConnectionChannel.h.
 *
 * This implementation uses a Poco stream socket.
 */
class SocketForwardingChannel : public FaspConnectionChannel {

private:
    Poco::Net::StreamSocket _remoteFaspstreamSocket;
        
public:

    SocketForwardingChannel(int port) 
    {
        /*
         * Wait for client to do normal connection
         */
        Poco::Net::ServerSocket srv(port);
        _remoteFaspstreamSocket = srv.acceptConnection();
        /*
         * Ensure a timeout is set on the socket. This will be
         * used to check when all data from the remote has been
         * read, and it is waiting for a response.
         */
        _remoteFaspstreamSocket.setReceiveTimeout(Poco::Timespan(100000));
    }

    int write(const char *data, size_t size)
    {
        /*
         * Try to write data to Poco stream socket (connected to
         * remote client).
         */
        try {
            return _remoteFaspstreamSocket.sendBytes(data, size);
        }
        catch (Poco::Net::ConnectionResetException&) {
            /*
             * Connection to remote client closed unexpectedly.
             */
            throw FaspChannelClosedException();
        }
    }

    int read(char* buffer, size_t size) {
        /*
         * Try to read from remote client
         */
        try {
            return _remoteFaspstreamSocket.receiveBytes(buffer, size);
        }
        catch (Poco::TimeoutException&) {
            /*
             * Timed out reading from remote client; we assume they aren't
             * trying to send anything, and are either busy or waiting for
             * a reply from the server.
             */
            throw FaspChannelTimeoutException();
        }
        catch (Poco::Net::ConnectionResetException&) {
            /*
             * Connection to remote client closed unexpectedly.
             */
            throw FaspChannelClosedException();
        }
    }

    void close() {
        /*
         * Close our remote connection.
         *
         * NOTE: You probably won't want to do this if you are
         * sharing the TCP connection with other parts of the
         * application. We close the socket here in this simple
         * example.
         */
        _remoteFaspstreamSocket.close();
    }
        
};

static
void
setupStreams() {
#ifdef WIN32
    /*
     * Required on Windows for binary files
     */
    _setmode(fileno(stdin), O_BINARY);
    _setmode(fileno(stdout), O_BINARY);
#endif
    std::cin.exceptions( std::ifstream::badbit );
    std::cout.exceptions( std::ifstream::badbit );
}

static
void
log_msg(const char *msg)
{
    fprintf(stderr, "%s\n", msg);
}


int main()
{
    char data[512];
    setupStreams();
    try {
        /*
         * Set up regular TCP connection with client.
         *
         * IMPORTANT: The destructor for your implementation of
         *            FaspConnectionChannel must fire *after* the
         *            destructor for the FaspStreamServer. Otherwise,
         *            you may get a seg fault in the background thread.
         *
         *            This example will work properly because destructors
         *            for variables on the stack are fired in the reverse
         *            order they are allocated, in C++.
         */
        std::cerr << "Waiting for client to connect...\n";
        SocketForwardingChannel sf(4444);
        std::cerr << "Client connected, setting up faspstream tunneling...\n";
        /*
         * Initialize faspstream server
         */
        FaspStreamServer fss("Unique-ID-For-a-Client", 
            dynamic_cast<FaspConnectionChannel *>(&sf));
        /*
         * Configure server
         * ----------------
         *
         * Set target rate to 100 Mbps
         */
        fss.setRateTarget(100000);
        /*
         * Set the port to connect to
         */
        fss.setPort(4444);
        /*
         * Set the UDP Port for the fasp transfer
         */
        fss.setUdpPort(33001);

        fss.registerLogCallback(log_msg);

        fss.setTransferMode(FASPSTREAM_TRANSFER_MODE_TYPE_DIRECT_RECEIVE);
        /*
         * Start listening for client
         * --------------------------
         */
        fss.startServer(true);
        std::cerr << "Tunneling set up, receiving data..." << std::endl;
        /*
         * Receive data
         */

        long totalCount = 0;
        int count = 0;
        do {
            /*
             * Read from fasp stream
             */
            count = fss.getFaspInputStream().read(data, 0, sizeof(data));
            totalCount += count;
            /*
             * Write to stdout
             */
            std::cout.write(data, count);

            std::cerr << "Bytes Received: " << totalCount << "\r";
            std::cerr.flush();
        } while (count > 0);
        
        std::cout.flush();
        /*
         * Clean up server
         */
        fss.cleanup();
        std::cerr << std::endl << "Transfer Complete!" << std::endl;
        std::cerr << "Bytes Received: " << totalCount << std::endl;
        /*
         * Tear down socket channel
         */
        sf.close();
    } 
    catch (std::ofstream::failure &e) {
        std::cerr << "Error writing to stdout: " << std::endl;
        std::cerr << e.what();
    }
    catch (std::exception &e) {
        std::cerr << "A standard exception occurred:" << std::endl;
        std::cerr << e.what();
    }
    catch (...) {
        std::cerr << "An unknown exception ocurred! ";
        std::cerr << "(No further information is available)";
    }

    return 0;
}

